from typing import Optional, Union

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None
try:
    from geopandas import GeoDataFrame
except ImportError:
    GeoDataFrame = None


def normalize_data(
    data: Optional[Union[str, DataFrame, GeoDataFrame]]
) -> Optional[str]:
    """Prepare user-provided data for transport to JS.

    Args:
        data: CSV/JSON String data or DataFrame to prepare for sending to Studio
    """
    if data is None:
        return None

    if isinstance(data, str):
        return data

    if GeoDataFrame is not None and isinstance(data, GeoDataFrame):
        # Reproject to 4326 if needed
        if data.crs and getattr(data.crs, 'to_epsg', lambda: 0)() != 4326:
            data = data.to_crs(epsg=4326)

    return data.to_csv()



import json
from asyncio import Future
from typing import Any, Callable, Dict, List, Optional, Sequence, Type, Union
from uuid import UUID, uuid4

from deprecated import deprecated
from ipywidgets import DOMWidget, Widget
from pydantic import BaseModel
from traitlets import Bool, Int, Unicode
from traitlets import Union as TraitletsUnion

from . import models
from ._frontend import module_name, module_version
from .models import Action


class UnfoldedMap(DOMWidget):
    """TODO: Add docstring here"""

    _model_name = Unicode('UnfoldedMapModel').tag(sync=True)
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)
    _view_name = Unicode('UnfoldedMapView').tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)

    value = Unicode('Hello from Jupyter').tag(sync=True)

    # TODO: support all of MapOptions arguments?
    mapUrl = Unicode('').tag(sync=True)
    mapUUID = Unicode('').tag(sync=True)
    width = TraitletsUnion(
        [Unicode(), Int()], allow_none=True, default_value='100%'
    ).tag(sync=True)
    height = TraitletsUnion(
        [Unicode(), Int()], allow_none=True, default_value='100%'
    ).tag(sync=True)
    iframe = Bool(True).tag(sync=True)
    sdkUrl = Unicode('').tag(sync=True)

    def __init__(
        self,
        on_load: models.OnLoadHandlerType = None,
        on_timeline_interval_change: models.OnTimelineIntervalChangeHandlerType = None,
        on_layer_timeline_time_change: models.OnLayerTimelineTimeChangeHandlerType = None,
        on_filter: models.OnFilterHandlerType = None,
        **kwargs: Any,
    ):
        super().__init__(**kwargs)

        # Register callback for receiving messages
        self.on_msg(self._receive_message)

        # Mapping from message id to related future
        self.futures: Dict[str, Dict] = {}
        self.map_event_handlers = models.MapEventHandlers(
            on_load=on_load,
            on_timeline_interval_change=on_timeline_interval_change,
            on_layer_timeline_time_change=on_layer_timeline_time_change,
            on_filter=on_filter,
        )

    def _send_message(self, message: Action) -> None:
        """Send message to JS

        TODO: should this optionally poll for a response synchronously?
        TODO: update to handle sending binary buffers
        """
        # In order to properly serialize the Pydantic model to a dict, this
        # serializes to JSON and then loads it back as a dict. Currently,
        # exporting a Pydantic model to a dict does not make the contained
        # objects serializable. (https://stackoverflow.com/q/65622045). This
        # means that later calling `json.dumps` on this dict will fail.
        # On the other hand, ipywidgets requires dict input so it can do its own
        # JSON serialization.
        # https://github.com/jupyter-widgets/ipywidgets/blob/24628f006c8a468994eba7a6e964fe619c992267/ipywidgets/widgets/widget.py#L516-L526
        msg = json.loads(message.json())
        self.send(msg)
        # print('sent message:', msg)

    def _create_future(self, message: Action, response_callback: Callable) -> Future:
        future: Future = Future()
        self.futures[str(message.message_id)] = {
            'future': future,
            'response_callback': response_callback,
        }
        return future

    # TODO: check type of `buffers`
    def _receive_message(
        self, widget: Widget, content: Dict, buffers: List[bytes]
    ) -> None:
        """Receive message from JS"""
        # pylint:disable=unused-argument
        if 'eventType' in content:
            return self._receive_event(content)

        if 'messageId' in content:
            return self._receive_message_response(content)

    def _receive_message_response(self, content: Dict) -> None:
        """Receive response message from JS"""
        message_id = content['messageId']
        data = content['data']
        self.received_message = content

        future_ref = self.futures.pop(message_id)
        response_callback = future_ref['response_callback']
        self.future_ref = future_ref
        future_ref['future'].set_result(response_callback(data))

    def _receive_event(self, content: Dict) -> None:
        """Receive event notification from JS"""
        event_type = content['eventType']
        data = content['data']
        self.value = event_type
        callback = getattr(self.map_event_handlers, event_type)
        if callback:
            callback(data)

    # TODO generic typing?
    # TODO: error responses have a data object of type {message: '...'}
    @staticmethod
    def _create_iterable_response_callback(
        response_class: Type[BaseModel],
    ) -> Callable[[List[Dict]], List[BaseModel]]:
        def response_callback(resp: List[Dict]) -> List[BaseModel]:
            if resp:
                try:
                    return [response_class(**item) for item in resp]
                except:
                    pass

            return []

        return response_callback

    # TODO generic typing?
    @staticmethod
    def _create_response_callback(
        response_class: Type[BaseModel],
    ) -> Callable[[Dict], Optional[BaseModel]]:
        def response_callback(resp: Dict) -> Optional[BaseModel]:
            if resp:
                try:
                    return response_class(**resp)
                except:
                    pass

            # TODO: what to return as default?
            return None

        return response_callback

    @staticmethod
    def get_map_url(map_uuid: Union[UUID, str]) -> str:
        """Get the full URL for the published map

        Args:
            map_uuid - Universally unique identifier (UUID) for the published map

        Returns:
            (string): Full URL for the published map
        """
        return f'https://studio.unfolded.ai/public/{map_uuid}'

    # TODO: return future types
    def set_view_state(self, view_state: models.ViewState) -> Future:
        """Set the map view state

        Args:
            view_state: ViewState model instance or dict with `longitude`, `latitude`, and `zoom` keys.
        """
        action = models.SetViewState(view_state=view_state)
        self._send_message(action)
        response_callback = self._create_response_callback(models.ViewState)
        return self._create_future(action, response_callback)

    def get_layers(self) -> Future:
        """Get all layers for the provided map instance"""
        action = models.GetLayers()
        self._send_message(action)
        response_callback = self._create_iterable_response_callback(models.Layer)
        return self._create_future(action, response_callback)

    def set_layer_visibility(self, layer_id: str, is_visible: bool) -> Future:
        """Set visibility of specified layer

        Args:
            layer_id: layer id
            is_visible: If True, make layer visible, else hide the layer
        """
        action = models.SetLayerVisibility(layer_id=layer_id, is_visible=is_visible)
        self._send_message(action)
        response_callback = self._create_response_callback(models.Layer)
        return self._create_future(action, response_callback)

    def set_theme(self, theme: str) -> None:
        """Set the map theme to 'light' or 'dark'

        Args:
            theme: theme name, either 'light' or 'dark'
        """
        action = models.SetTheme(theme=theme)
        self._send_message(action)

    def get_timeline_info(self, idx: str) -> Future:
        """Get information object for the timeline filter

        Args:
            idx: Index of the timeline filter
        """
        action = models.GetTimelineInfo(idx=idx)
        self._send_message(action)
        response_callback = self._create_response_callback(models.TimelineInfo)
        return self._create_future(action, response_callback)

    @deprecated(reason="Use `setTimelineConfig` instead")
    def toggle_timeline_animation(self, idx: str) -> None:
        """Toggle timeline filter animation

        Args:
            idx: Index of the timeline filter
        """
        action = models.ToggleTimelineAnimation(idx=idx)
        self._send_message(action)

    @deprecated(reason="Use `setTimelineConfig` instead")
    def toggle_timeline_visibility(self, idx: str) -> None:
        """Toggle timeline filter visibility

        Args:
            idx: Index of the timeline filter
        """
        action = models.ToggleTimelineVisibility(idx=idx)
        self._send_message(action)

    @deprecated(reason="Use `setTimelineConfig` instead")
    def set_timeline_interval(
        self, idx: str, start_time: float, end_time: float
    ) -> None:
        """Set current timeline filter interval

        Args:
            idx: Index of the timeline filter
            start_time: Unix Time in milliseconds for the start of the interval
            end_time: Unix Time in milliseconds for the end of the interval
        """
        action = models.SetTimelineInterval(
            idx=idx, start_time=start_time, end_time=end_time
        )
        self._send_message(action)

    @deprecated(reason="Use `setTimelineConfig` instead")
    def set_timeline_animation_speed(self, idx: str, speed: float) -> None:
        """Set current time filter animation speed

        Args:
            idx: Index of the timeline filter
            speed: speed multiplier
        """
        action = models.SetTimelineAnimationSpeed(idx=idx, speed=speed)
        self._send_message(action)

    def set_timeline_config(self, config: models.TimelineConfig) -> None:
        """Set timeline configuration

        Args:
            config: Timeline configuration object
        """
        action = models.SetTimelineConfig(config=config)
        self._send_message(action)

    def refresh_map_data(self) -> None:
        """Refresh map data sources"""
        action = models.RefreshMapData()
        self._send_message(action)

    def get_layer_timeline_info(self) -> Future:
        """Get information object for the layer timeline control"""
        action = models.GetLayerTimelineInfo()
        self._send_message(action)
        response_callback = self._create_response_callback(models.LayerTimelineInfo)
        return self._create_future(action, response_callback)

    def set_layer_timeline_config(self, config: models.LayerTimelineConfig) -> Future:
        """Set layer timeline configuration

        Args:
            config: Layer timeline configuration object
        """
        action = models.SetLayerTimelineConfig(config=config)
        self._send_message(action)
        response_callback = self._create_response_callback(models.LayerTimelineConfig)
        return self._create_future(action, response_callback)

    def _add_tileset(self, tileset: models.Tileset) -> None:
        """Create a new tileset

        Args:
            tileset: tileset configuration
        """
        action = models.AddTileset(tileset=tileset)
        self._send_message(action)

    def add_dataset(
        self, dataset: Union[UUID, str, models.Dataset], auto_create_layers: bool = True
    ) -> Future:
        """Add a dataset to the map.
        If a dataset with data is provided, a new dataset will be created and added to the map (but not uploaded to Studio).
        Alternatively, if just a uuid is provided or a Dataset with no data specified in it, the method will attempt to add
        a previously uploaded dataset with the specified uuid to the map.

        Args:
            dataset: Dataset for creating and adding a new dataset to the map or a uuid of a previously uploaded dataset.
            auto_create_layers: When `True` Studio will attempt to create layers automatically
        """
        if isinstance(dataset, UUID) or isinstance(dataset, str):
            dataset = models.Dataset(uuid=dataset)
        action = models.AddDataset(
            dataset=dataset, auto_create_layers=auto_create_layers
        )
        response_callback = self._create_response_callback(models.AddDatasetResponse)
        self._send_message(action)
        return self._create_future(action, response_callback)

    def remove_dataset(self, uuid: UUID) -> Future:
        """Remove the dataset with the specified UUID from the map.

        Args:
            uuid: Dataset UUID
        """
        action = models.RemoveDataset(uuid=uuid)
        self._send_message(action)
        response_callback = self._create_response_callback(models.RemoveDatasetResponse)
        return self._create_future(action, response_callback)

    def add_layer(self, layer: models.LayerSpec) -> None:
        """Add a layer to the map

        Args:
            layer: Layer configuration
        """
        action = models.AddLayer(layer=layer)
        self._send_message(action)

    def remove_layer(self, layer_id: str) -> None:
        """Remove layer from the map

        Args:
            id: Layer id
        """
        action = models.RemoveLayer(id=layer_id)
        self._send_message(action)

    def set_filter(self, info: models.FilterInfo) -> None:
        """Set filter value

        Args:
            info: Filter info to set
        """
        action = models.SetFilter(info=info)
        self._send_message(action)

    def remove_filter(self, filter_id: str) -> None:
        """Remove filter from the map

        Args:
            id: Filter id
        """
        action = models.RemoveFilter(id=filter_id)
        self._send_message(action)

    def set_map_event_handlers(
        self, event_handlers: Optional[models.MapEventHandlers]
    ) -> None:
        """Sets event handlers for the specified map to get notifications.
           Only one event handler per message type can be registered,
           so subsequent calls will override previously set event handler
           for the specified message types.
           This method only updates callbacks for those message types
           which are passed. The others will remain unchanged.
           Set specific event handlers to None to unregister them,
           or pass None to the method to remove all handlers.
        Args:
            map_event_handlers: MapEventHandlers
        """
        if event_handlers is None:
            self.map_event_handlers = models.MapEventHandlers()
        else:
            update = (
                event_handlers.dict()
                if isinstance(event_handlers, models.MapEventHandlers)
                else event_handlers
            )
            self.map_event_handlers = self.map_event_handlers.copy(update=update)


class HTMLUnfoldedMap(UnfoldedMap):
    datasets: List[models.Dataset]
    layers: List[models.LayerSpec]

    def __init__(
        self,
        *,
        datasets: Sequence[Union[models.Dataset, Dict]] = (),
        layers: Sequence[Union[models.LayerSpec, Dict]] = (),
        **kwargs: Any,
    ):
        super().__init__(**kwargs)

        dataset_models: List[models.Dataset] = [
            models.Dataset(**dataset)
            if not isinstance(dataset, models.Dataset)
            else dataset
            for dataset in datasets
        ]
        layer_models: List[models.LayerSpec] = [
            models.LayerSpec(**layer)
            if not isinstance(layer, models.LayerSpec)
            else layer
            for layer in layers
        ]

        self.datasets = dataset_models
        self.layers = layer_models

    def _repr_html_(self):
        # Per-iframe identifier
        frame_uuid = uuid4()
        if self.mapUrl:
            url = self.mapUrl
        elif self.mapUUID:
            url = 'https://studio.unfolded.ai/map/' + self.mapUUID
        else:
            url = 'https://studio.unfolded.ai/incognito'

        # NOTE: in f-strings, double quotes '{{' and '}}' resolve to a literal '{', '}'
        # Begin script
        html = f"""
        <script>(function() {{
        var frame;
        var queue = [];
        var lastSentType;
        window.addEventListener('message', function(evt) {{
          var data = `${{evt.data}}`;
          if (data.indexOf('map-sdk-initialized-response') >= 0) onInit();
          if (lastSentType && data.indexOf(`${{lastSentType}}-response` >= 0)) sendNext();
        }});
        function onInit() {{
          frame = document.getElementById('{frame_uuid}');
        """

        # Add datasets
        html += self._repr_html_datasets()

        # Add layers
        html += self._repr_html_layers()

        # End script
        html += """
            sendNext();
        }
        function sendNext() {
          if (queue.length === 0) {
            lastSentType = null;
            return;
          }
          var msg = queue.shift();
          if (frame) {
              frame.contentWindow.postMessage(JSON.stringify(msg), '*');
              lastSentType = msg.type;
          }
        }
        })();</script>
        """

        # The default 100% height leads to the iframe height
        # growing infinitely in Databricks due to their resizing logic
        height = self.height
        try:
            int(self.height)
        except ValueError:
            height = 400

        iframe_tag = f"""
        <iframe id="{frame_uuid}" src="{url}" width="{self.width}" height="{height}" />
        """

        return html + iframe_tag

    def _repr_html_datasets(self) -> str:
        """Helper to create HTML representation of self.datasets"""
        if not self.datasets:
            return ''

        dataset_str = """
          queue.push({{
            type: 'map-sdk-add-dataset-to-map',
            source: 'unfolded-sdk-client',
            data: {{ autoCreateLayers: {auto_layers}, dataset: {dataset} }}
          }});
        """
        auto_layers = 'false' if self.layers else 'true'

        encoded = [
            dataset_str.format(
                auto_layers=auto_layers, dataset=dataset.json(by_alias=True)
            )
            for dataset in self.datasets
        ]
        return '\n'.join(encoded)

    def _repr_html_layers(self) -> str:
        """Helper to create HTML representation of self.layers"""
        if not self.layers:
            return ''

        layer_str = """
          queue.push({{
            type: 'map-sdk-add-layer',
            source: 'unfolded-sdk-client',
            data: {{layer: {layer}}}
          }});
        """
        encoded = [
            layer_str.format(layer=layer.json(by_alias=True)) for layer in self.layers
        ]
        return '\n'.join(encoded)



def test_nbextension_path() -> None:
    # Check that magic function can be imported from package root:
    # pylint:disable=import-outside-toplevel
    from unfolded.map_sdk import _jupyter_nbextension_paths

    # Ensure that it can be called without incident:
    path = _jupyter_nbextension_paths()
    # Some sanity checks:
    assert len(path) == 1
    assert isinstance(path[0], dict)



import sys
from typing import Any, Callable, Dict, List, Optional, Tuple, Union
from uuid import UUID, uuid4

from pydantic import BaseModel, Field, validator

from .enums import ActionType, FilterType, LayerType
from .transfer_utils import normalize_data

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None

# Literal is included in the stdlib as of Python 3.8
if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal


class CamelCaseBaseModel(BaseModel):
    class Config:
        allow_population_by_field_name = True

        @staticmethod
        def to_camel_case(snake_str: str) -> str:
            """Convert snake_case string to camelCase
            https://stackoverflow.com/a/19053800
            """
            components = snake_str.split('_')
            # We capitalize the first letter of each component except the first one
            # with the 'title' method and join them together.
            return components[0] + ''.join(x.title() for x in components[1:])

        alias_generator = to_camel_case


# TODO: do we expose this class at all?
# class MapOptions(BaseModel):
#     mapUrl: Optional[str]
#     mapUUID: Optional[UUID]
#     appendToId: Optional[str]
#     id: Optional[str]
#     embed: Optional[bool]
#     onLoad: Optional[() => void]
#     onTimelineIntervalChange: Optional[(currentTimeInterval: List[float>) => void]
#     onLayerTimelineTimeChange: Optional[(currentTime: float) => void]
#     appendToDocument: bool
#     width: float
#     height: float

# class MapInstance(BaseModel):
#     iframe: HTMLCanvasElement

FilterValueType = Union[List[float], List[str], bool, Any]


class FilterChangeEvent(CamelCaseBaseModel):
    """Information on filter changes passed to onFilter callbacks

    Kwargs:
        id: Filter ID
        name: Field names the filter applies to
        data_id: Dataset UUID
        prop: Filter property that is changed: 'value', 'name', etc
        type: FilterType
        value: Value of the filter
    """

    id: str
    name: List[str]
    data_id: UUID
    prop: str
    type: FilterType
    value: FilterValueType


class ClickEvent(CamelCaseBaseModel):
    """Information on the object passed to on_click

    Kwargs:
        index: Data row index
        coordinate: Geospatial coordinates [lng, lat]
        x: Mouse position x relative to the viewport
        y: Mouse position y relative to the viewport
        layer_id: The layer id that the picked object belongs to
        data: Picked object in the layer
    """

    index: int
    coordinate: Tuple[float, float]
    x: float
    y: float
    layer_id: Optional[str]
    data: Optional[Dict[str, Any]]


class HoverEvent(CamelCaseBaseModel):
    """Information on the object passed to on_hover

    Kwargs:
        index: Data row index
        coordinate: Geospatial coordinates [lng, lat]
        x: Mouse position x relative to the viewport
        y: Mouse position y relative to the viewport
        layer_id: The layer id that the picked object belongs to
        data: Picked object in the layer
    """

    index: int
    coordinate: Tuple[float, float]
    x: float
    y: float
    layer_id: Optional[str]
    data: Optional[Dict[str, Any]]


class Geometry(CamelCaseBaseModel):
    """GeoJSON geometry object`"""

    type: str
    coordinates: Any


class Feature(CamelCaseBaseModel):
    """GeoJSON feature object`"""

    id: str
    properties: Any
    geometry: Geometry


class GeometrySelectionEvent(CamelCaseBaseModel):
    """Information on the object passed to on_geometry_selection

    Kwargs:
        features: List of drawn GeoJSON shapes
    """

    features: List[Feature]


OnLoadHandlerType = Optional[Callable[[None], None]]
OnTimelineIntervalChangeHandlerType = Optional[Callable[[List[int]], None]]
OnLayerTimelineTimeChangeHandlerType = Optional[Callable[[int], None]]
OnFilterHandlerType = Optional[Callable[[FilterChangeEvent], None]]
OnHoverHandlerType = Optional[Callable[[HoverEvent], None]]
OnClickHandlerType = Optional[Callable[[ClickEvent], None]]
OnGeometrySelectionHandlerType = Optional[Callable[[GeometrySelectionEvent], None]]


class MapEventHandlers(CamelCaseBaseModel):
    """Event handlers which can be registered for receiving notifications of certain map events."""

    on_load: OnLoadHandlerType
    on_timeline_interval_change: OnTimelineIntervalChangeHandlerType
    on_layer_timeline_time_change: OnLayerTimelineTimeChangeHandlerType
    on_filter: OnFilterHandlerType
    on_hover: OnHoverHandlerType
    on_click: OnClickHandlerType
    on_geometry_selection: OnGeometrySelectionHandlerType


class Layer(CamelCaseBaseModel):
    """Layer info returned by `get_layers()`"""

    label: str
    id: str
    is_visible: bool


class ViewState(CamelCaseBaseModel):
    """Map viewport state for use with `set_view_state()`"""

    longitude: float
    latitude: float
    # TODO: From looking at the SDK implementation, this should be Optional?
    zoom: Optional[float] = None


class TimelineInfo(CamelCaseBaseModel):
    """Timeline control state returned by `get_timeline_info()`"""

    data_id: List[UUID]
    domain: List[float]
    is_visible: bool
    enlarged_histogram: List[Any]
    histogram: List[Any]
    value: List[float]
    speed: float
    step: float
    is_animating: bool


class TimeInterval(CamelCaseBaseModel):
    """Time interval for use with `set_timeline_config()`"""

    start_time: float
    end_time: float


class TimelineConfig(CamelCaseBaseModel):
    """Time configuration for use with `set_timeline_config()`"""

    idx: int
    current_time_interval: Optional[TimeInterval]
    is_visible: Optional[bool]
    is_animating: Optional[bool]
    speed: Optional[float]
    timezone: Optional[str]
    time_format: Optional[str]


class LayerTimelineInfo(CamelCaseBaseModel):
    """Layer timeline state returned by `get_layer_timeline_info()`"""

    current_time: float
    default_time_format: str
    domain: List[float]
    duration: float
    is_visible: bool
    is_animating: bool
    speed: float
    time_format: str
    time_steps: Any
    timezone: str


class LayerTimelineConfig(CamelCaseBaseModel):
    """Layer timeline configuration for use with `set_layer_timeline_config()`"""

    current_time: Optional[float]
    is_visible: Optional[bool]
    is_animating: Optional[bool]
    speed: Optional[float]
    timezone: Optional[str]
    time_format: Optional[str]


TilesetType = Literal['vector-tile', 'raster-tile']


class VectorMeta(CamelCaseBaseModel):
    """Vector tileset metadata"""

    data_url: str
    metadata_url: str


class RasterMeta(CamelCaseBaseModel):
    """Raster tileset metadata"""

    data_url: str
    metadata_url: str


class Tileset(CamelCaseBaseModel):
    """Tileset configuration for use with `add_tileset`"""

    name: str
    type: TilesetType
    meta: Union[VectorMeta, RasterMeta]


class LayerConfig(CamelCaseBaseModel):
    """Layer configuration for use with `add_layer` or `add_dataset`.
    `LayerConfig` is part of `LayerSpec`.
    """

    data_id: UUID
    label: Optional[str]
    columns: Dict[str, Any]
    is_visible: Optional[bool]
    vis_config: Optional[Dict[str, Any]]
    color_field: Optional[Dict[str, Any]]
    color_scale: Optional[str]
    color_ui: Optional[Dict[str, Any]] = Field(alias='colorUI')


class LayerSpec(CamelCaseBaseModel):
    """Layer configuration for use with `add_layer` or `add_dataset`"""

    id: str
    type: LayerType
    config: LayerConfig
    visual_channels: Optional[Dict[str, Any]]


class Dataset(CamelCaseBaseModel):
    """Dataset configuration for use with `add_dataset`.
    Kwargs:
        uuid: Unique identifier of the dataset (will be auto-generated when not provided).
        data: (Optional) CSV, JSON, DataFrame or GeoDataFrame data for the dataset.
    """

    uuid: UUID = Field(default_factory=uuid4)
    label: str = 'Untitled'
    data: Optional[Union[str, DataFrame]]

    @validator('data', pre=True, always=True)
    # pylint: disable=no-self-argument
    def set_data_normalized(
        cls, data: Optional[Union[str, DataFrame]]
    ) -> Optional[str]:
        return normalize_data(data)

    class Config:
        # to prevent "No validator for DataFrame"
        arbitrary_types_allowed = True
        validate_assignment = True


class FilterInfo(CamelCaseBaseModel):
    """Filter settings for use with `set_filter`"""

    id: str
    data_id: Optional[UUID]
    field: str
    value: FilterValueType


class AddDatasetResponse(BaseModel):
    """Response of `add_dataset`calls"""

    status: Literal['ok', 'fail']
    message: Optional[str]


class RemoveDatasetResponse(BaseModel):
    """Response of `remove_dataset`calls"""

    status: Literal['ok', 'fail']
    message: Optional[str]


class Action(CamelCaseBaseModel):
    """Base Action payload class"""

    type: ActionType
    message_id: UUID = Field(default_factory=uuid4)

    def dict(self, **kwargs: Any) -> Dict[str, Any]:
        """Custom dict encoding to put keys within `data` key

        Only top level keys should be type, data, and messageId
        """
        # Export by alias by default, but don't change by_alias=False
        if 'by_alias' not in kwargs.keys():
            kwargs['by_alias'] = True

        # Remove None values by default
        if 'exclude_none' not in kwargs.keys():
            kwargs['exclude_none'] = True

        # Use default superclass dict serialization
        d = super().dict(**kwargs)

        # Make sure the data key doesn't exist yet
        assert d.get('data') == None, 'data key already exists'

        d['data'] = {}

        # Keys that should stay at the top level
        top_level_json_keys = ['type', 'messageId', 'data']

        # Set all keys within the data key, but not recursively
        inner_keys = set(d.keys()).difference(top_level_json_keys)
        for key in inner_keys:
            d['data'][key] = d.pop(key)

        return d

    def json(self, **kwargs: Any) -> str:
        """Custom json encoding to put keys within `data` key

        Only top level keys should be type, data, and messageId
        """
        # Export by alias by default, but don't change by_alias=False
        if 'separators' not in kwargs.keys():
            kwargs['separators'] = (',', ':')

        # Export by alias by default, but don't change by_alias=False
        if 'by_alias' not in kwargs.keys():
            kwargs['by_alias'] = True

        # Remove None values by default
        if 'exclude_none' not in kwargs.keys():
            kwargs['exclude_none'] = True

        return super().json(**kwargs)


class SetViewState(Action):
    """Action payload sent with `set_view_state`calls"""

    type: ActionType = ActionType.SET_VIEW_STATE
    view_state: ViewState


class GetLayers(Action):
    """Action payload sent with `get_layers`calls"""

    type: ActionType = ActionType.GET_LAYERS


class SetLayerVisibility(Action):
    """Action payload sent with `set_layer_visibility`calls"""

    type: ActionType = ActionType.SET_LAYER_VISIBILITY
    layer_id: str
    is_visible: bool


class SetTheme(Action):
    """Action payload sent with `set_theme`calls"""

    type: ActionType = ActionType.SET_THEME
    theme: Literal['light', 'dark']


class GetTimelineInfo(Action):
    """Action payload sent with `get_timeline_info`calls"""

    type: ActionType = ActionType.GET_TIMELINE_INFO
    idx: int


class ToggleTimelineAnimation(Action):
    """Action payload sent with `toggle_timeline_animation`calls"""

    type: ActionType = ActionType.TOGGLE_TIMELINE_ANIMATION
    idx: int


class ToggleTimelineVisibility(Action):
    """Action payload sent with `toggle_timeline_visibility`calls"""

    type: ActionType = ActionType.TOGGLE_TIMELINE_VISIBILITY
    idx: int


class SetTimelineInterval(Action):
    """Action payload sent with `set_timeline_interval`calls"""

    type: ActionType = ActionType.SET_TIMELINE_INTERVAL
    idx: int
    start_time: float
    end_time: float


class SetTimelineAnimationSpeed(Action):
    """Action payload sent with `set_timeline_animation_speed`calls"""

    type: ActionType = ActionType.SET_TIMELINE_SPEED
    idx: int
    speed: float


class SetTimelineConfig(Action):
    """Action payload sent with `set_timeline_config` calls"""

    type: ActionType = ActionType.SET_TIMELINE_CONFIG
    config: TimelineConfig


class RefreshMapData(Action):
    """Action payload sent with `refresh_map_data` calls"""

    type: ActionType = ActionType.REFRESH_MAP_DATA


class GetLayerTimelineInfo(Action):
    """Action payload sent with `get_layer_timeline_info` calls"""

    type: ActionType = ActionType.GET_LAYER_TIMELINE_INFO


class SetLayerTimelineConfig(Action):
    """Action payload sent with `set_layer_timeline_config` calls"""

    type: ActionType = ActionType.SET_LAYER_TIMELINE_CONFIG
    config: LayerTimelineConfig


class AddTileset(Action):
    """Action payload sent with `add_tileset` calls"""

    type: ActionType = ActionType.ADD_TILESET_TO_MAP
    tileset: Tileset


class AddDataset(Action):
    """Action payload sent with `add_dataset` calls"""

    type: ActionType = ActionType.ADD_DATASET_TO_MAP
    dataset: Dataset
    auto_create_layers: bool


class RemoveDataset(Action):
    """Action payload sent with `remove_dataset` calls"""

    type: ActionType = ActionType.REMOVE_DATASET_FROM_MAP
    uuid: UUID


class AddLayer(Action):
    """Action payload sent with `add_layer` calls"""

    type: ActionType = ActionType.ADD_LAYER
    layer: LayerSpec


class RemoveLayer(Action):
    """Action payload sent with `remove_layer` calls"""

    type: ActionType = ActionType.REMOVE_LAYER
    id: str


class RemoveFilter(Action):
    """Action payload sent with `remove_filter` calls"""

    type: ActionType = ActionType.REMOVE_FILTER
    id: str


class SetFilter(Action):
    """Action payload sent with `set_filter` calls"""

    type: ActionType = ActionType.SET_FILTER
    info: FilterInfo



from typing import Dict, List

from ._version import __version__
from .unfolded_jupyter import HTMLUnfoldedMap, UnfoldedMap


def _jupyter_labextension_paths() -> List[Dict[str, str]]:
    """Called by Jupyter Lab Server to detect if it is a valid labextension and
    to install the widget
    Returns
    =======
    src: Source directory name to copy files from. Webpack outputs generated files
        into this directory and Jupyter Lab copies from this directory during
        widget installation
    dest: Destination directory name to install widget files to. Jupyter Lab copies
        from `src` directory into <jupyter path>/labextensions/<dest> directory
        during widget installation
    """
    return [{'src': 'labextension', 'dest': '@unfolded/jupyter-map-sdk'}]


def _jupyter_nbextension_paths() -> List[Dict[str, str]]:
    """Called by Jupyter Notebook Server to detect if it is a valid nbextension and
    to install the widget
    Returns
    =======
    section: The section of the Jupyter Notebook Server to change.
        Must be 'notebook' for widget extensions
    src: Source directory name to copy files from. Webpack outputs generated files
        into this directory and Jupyter Notebook copies from this directory during
        widget installation
    dest: Destination directory name to install widget files to. Jupyter Notebook copies
        from `src` directory into <jupyter path>/nbextensions/<dest> directory
        during widget installation
    require: Path to importable AMD Javascript module inside the
        <jupyter path>/nbextensions/<dest> directory
    """
    return [
        {
            'section': 'notebook',
            'src': 'nbextension',
            'dest': 'unfolded/map_sdk',
            'require': 'unfolded/map_sdk/extension',
        }
    ]



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function

import os
from glob import glob
from os.path import join as pjoin

from jupyter_packaging import (
    combine_commands,
    create_cmdclass,
    ensure_targets,
    get_version,
    install_npm,
)
from setuptools import find_namespace_packages, setup

HERE = os.path.dirname(os.path.abspath(__file__))

with open("README.md") as f:
    readme = f.read()

# The name of the project
name = 'unfolded.map-sdk'
# Directory as a tuple since we use namespace packaging
code_directory = ('unfolded', 'map_sdk')

# Get the version
version = get_version(pjoin(*code_directory, '_version.py'))

# Representative files that should exist after a successful build
jstargets = [
    pjoin(HERE, *code_directory, 'nbextension', 'index.js'),
    pjoin(HERE, 'lib', 'plugin.js'),
]

package_data_spec = {name: ['nbextension/**js*', 'labextension/**']}

data_files_spec = [
    (
        'share/jupyter/nbextensions/unfolded/map_sdk',
        pjoin(*code_directory, 'nbextension'),
        '**',
    ),
    # These are intended to be npm package name
    (
        'share/jupyter/labextensions/@unfolded/jupyter-map-sdk',
        pjoin(*code_directory, 'labextension'),
        '**',
    ),
    ('share/jupyter/labextensions/@unfolded/jupyter-map-sdk', '.', 'install.json'),
    ('etc/jupyter/nbconfig/notebook.d', '.', 'unfolded.map_sdk.json'),
]

cmdclass = create_cmdclass(
    'jsdeps', package_data_spec=package_data_spec, data_files_spec=data_files_spec
)
cmdclass['jsdeps'] = combine_commands(
    install_npm(HERE, build_cmd='build:prod', npm=['yarn']),
    ensure_targets(jstargets),
)

setup_args = dict(
    name=name,
    description='Jupyter Widget for Unfolded.ai Maps',
    long_description=readme,
    long_description_content_type='text/markdown',
    version=version,
    scripts=glob(pjoin('scripts', '*')),
    cmdclass=cmdclass,
    packages=find_namespace_packages(include=['unfolded.*']),
    author='UnfoldedInc.',
    author_email='info@unfolded.ai',
    license='Unlicensed',
    platforms='Linux, Mac OS X, Windows',
    keywords=['Jupyter', 'Widgets', 'IPython'],
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Framework :: Jupyter',
    ],
    include_package_data=True,
    python_requires='>=3.6',
    install_requires=[
        'deprecated',
        'ipywidgets>=7.0.0',
        'pydantic',
        'typing_extensions;python_version<"3.8"',
    ],
    extras_require={
        'test': [
            'pytest>=4.6',
            'pytest-cov',
            'nbval',
            'mypy==0.812',
            'jupyter_packaging',
        ],
        'geopandas': ['geopandas>=0.7.0', 'pandas'],
        'examples': [
            # Any requirements for the examples to run
        ],
        'docs': [
            'jupyter_sphinx',
            'nbsphinx',
            'nbsphinx-link',
            'pytest_check_links',
            'pypandoc',
            'recommonmark',
            'sphinx>=1.5',
            'sphinx_rtd_theme',
        ],
    },
    entry_points={},
)

if __name__ == '__main__':
    setup(**setup_args)



from enum import Enum


class ActionType(str, Enum):
    """Actions that can be passed to Studio"""

    ADD_TILESET_TO_MAP = 'map-sdk-add-tileset-top-map'
    ADD_DATASET_TO_MAP = 'map-sdk-add-dataset-to-map'
    REMOVE_DATASET_FROM_MAP = 'map-sdk-remove-dataset-from-map'
    GET_LAYER_TIMELINE_INFO = 'map-sdk-get-layer-timeline-info'
    ADD_LAYER = 'map-sdk-add-layer'
    REMOVE_FILTER = 'map-sdk-remove-filter'
    REMOVE_LAYER = 'map-sdk-remove-layer'
    GET_LAYERS = 'map-sdk-get-layers'
    GET_TIMELINE_INFO = 'map-sdk-get-timeline-info'
    LAYER_ANIMATION_TIME_CHANGED = 'map-sdk-layer-animation-time-changed'
    MAP_INITIALIZED = 'map-sdk-initialized'
    REFRESH_MAP_DATA = 'map-sdk-reload-map-data'
    SET_FILTER = 'map-sdk-set-filter'
    SET_LAYER_TIMELINE_CONFIG = 'map-sdk-set-layer-timeline-config'
    SET_LAYER_VISIBILITY = 'map-sdk-set-layer-visibility'
    SET_MAP_CONTROL_VISIBILITY = 'map-sdk-set-map-control-visibility'
    SET_THEME = 'map-sdk-set-theme'
    SET_TIMELINE_CONFIG = 'map-sdk-set-timeline-config'
    SET_TIMELINE_INTERVAL = 'map-sdk-set-timeline-interval'
    SET_TIMELINE_SPEED = 'map-sdk-set-timeline-speed'
    SET_VIEW_STATE = 'map-sdk-set-view-state'
    TIMELINE_INTERVAL_CHANGED = 'map-sdk-timeline-interval-changed'
    TOGGLE_TIMELINE_ANIMATION = 'map-sdk-toggle-timeline-animation'
    TOGGLE_TIMELINE_VISIBILITY = 'map-sdk-toggle-timeline-visibility'


class LayerType(str, Enum):
    POINT = 'point'
    ARC = 'arc'
    LINE = 'line'
    GRID = 'grid'
    HEXAGON = 'hexagon'
    GEOJSON = 'geojson'
    CLUSTER = 'cluster'
    ICON = 'icon'
    HEATMAP = 'heatmap'
    HEXAGONID = 'hexagonId'
    _3D = '3D'
    TRIP = 'trip'
    S2 = 's2'


class FilterType(str, Enum):
    RANGE = 'range'
    SELECT = 'select'
    INPUT = 'input'
    TIME_RANGE = 'timeRange'
    MULTI_SELECT = 'multiSelect'
    POLYGON = 'polygon'
